/*
 * console.c
 */

__asm("	text\n	dc.b	'$Id: console.c,v 1.32 1998/11/03 15:24:14 sharl Exp $'\n	even\n");

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <conio.h>
#include <mbctype.h>
#include <sys/socket.h>
#define __IOCS_INLINE__
#include <iocslib.h>
#define __DOS_INLINE__
#include <doslib.h>

#include "config.h"
#include "irc.h"
#include "console.h"
#include "command.h"
#include "channel.h"
#include "util.h"
#include "kanji.h"

#define	BDCOLOR		(0x0000)
#define	FGCOLOR		(0xFFFE)

#define TIMEOUT_APPEAR	(180)
#define TIMEOUT_COLUMN	(2)
#define CHANNEL_COLUMN	(8)
#define MODELINESIZE	(96)

time_t timeout;		/* inactive timeout */
static char modeline[MODELINESIZE + 1];

static int oldbrk;
static BOOLEAN ctrlx = FALSE;
static BOOLEAN timebeep = FALSE;

/****************************************
	ʁEJ[\ʒuۑEA
****************************************/
#define PLAINSIZE	(0x20000 / sizeof(unsigned short))
static unsigned short plain0[PLAINSIZE], plain1[PLAINSIZE], plain2[PLAINSIZE];
#define CURX	(0x0974)
#define CURY	(0x0976)
static unsigned short oldcurx, oldcury;

static void
savelayout()
{
	long ssp;
	unsigned short *vp;
	int i;

	ssp = _iocs_b_super(0);
	for (i = 0, vp = (unsigned short *)0xe00000; vp < (unsigned short *)0xe20000; i++, vp++) {
		plain0[i] = *(vp + PLAINSIZE * 0);
		plain1[i] = *(vp + PLAINSIZE * 1);
		plain2[i] = *(vp + PLAINSIZE * 2);
	}
	oldcurx = *(unsigned short *)CURX;
	oldcury = *(unsigned short *)CURY;
	_iocs_b_super(ssp);
}

static void
restorelayout()
{
	long ssp;
	unsigned short *vp;
	int i;

	ssp = _iocs_b_super(0);
	for (i = 0, vp = (unsigned short *)0xe00000; vp < (unsigned short *)0xe20000; i++, vp++) {
		*(vp + PLAINSIZE * 0) = plain0[i];
		*(vp + PLAINSIZE * 1) = plain1[i];
		*(vp + PLAINSIZE * 2) = plain2[i];
	}
	*(unsigned short *)CURX = oldcurx;
	*(unsigned short *)CURY = oldcury;
	_iocs_b_super(ssp);
}

/****************************************
	J[R[h𓾂
****************************************/
static int
getpcode(char *ep)
{
	int r, g, b;

	if (sscanf(ep, " %u %u %u", &r, &g, &b) != 3)
		return 0;
	return (g << 11) | (r << 6) | (b << 1);
}

/****************************************
	J[ݒ
****************************************/
static void
setcolortable()
{
	char *ep;

	ep = getenv("BDCOLOR");
	_iocs_tpalet(0, ep ? getpcode(ep) : BDCOLOR);
	ep = getenv("FGCOLOR");
	_iocs_tpalet(3, ep ? getpcode(ep) : FGCOLOR);
}

/****************************************
	J[\`
****************************************/
static void
curput(int x, int y)
{
	int i;
	int plain;
	static struct _fntbuf cursor = {
		8, 16,
	};

	if (x > CONSOLE_WIDTH - 1)
		return;
	for (plain = 1; plain < 3; plain++) {
		_iocs_tcolor(plain);
		_iocs_textget(x * 8, y * 16, &cursor);
		for (i = 12; i < cursor.yl; i++) {
			cursor.buffer[i] ^= 0xff;
		}
		_iocs_textput(x * 8, y * 16, &cursor);
	}
	_iocs_tcolor(0);
}

/****************************************
	modeline update
****************************************/
static void
mdupdate()
{
	char buf[256];
	time_t t;

	memset(modeline, '=', MODELINESIZE);
	*(modeline + MODELINESIZE) = '\0';
	if (irc_timeout > 0) {
	t = timeout + irc_timeout - time(NULL);
		if (t < TIMEOUT_APPEAR) {
		   	if (timebeep == FALSE) {
				_dos_putchar('\a');
		   		timebeep = TRUE;
		   	}
			sprintf(buf, " %d ", t);
			memcpy(modeline + TIMEOUT_COLUMN, buf, strlen(buf));
		}
	}
	if (current_channel >= 0 && chanlist[current_channel].active != FALSE) {
		CHANSTAT *ccsp = &chanlist[current_channel];
		char *p = modeline + CHANNEL_COLUMN;

		*p++ = ' ';
		if (ccsp->isoper)
			*p++ = '@';
		memcpy(p, ccsp->name, strlen(ccsp->name));
		p += strlen(ccsp->name);
		*p = ' ';
		if (ccsp->topic && *ccsp->topic) {
			size_t len = strlen(ccsp->topic);

			p += 3;
			*p++ = ' ';
			if (p - modeline + len > MODELINESIZE)
				len = MODELINESIZE - (p - modeline) - 1;
			memcpy(p, ccsp->topic, len);
			p += len;
			*p = ' ';
		}
	}

	_iocs_b_putmes(13, 0, CONSOLE_HEIGHT, CONSOLE_WIDTH - 1, modeline);
}

/****************************************
	s͕ɕ\
****************************************/
static void
mdset(char *mes)
{
	_iocs_b_putmes(3, 0, CONSOLE_INPUT, CONSOLE_WIDTH - 1, mes);
}

/****************************************
	ʏ
****************************************/
void
init_console(BOOLEAN init)
{
	memset(modeline, '=', MODELINESIZE);
	*(modeline + MODELINESIZE) = '\0';
	_iocs_crtmod(0x10);
	setcolortable();
	_iocs_b_consol(0, 0, CONSOLE_WIDTH - 1, CONSOLE_HEIGHT - 1);
	_iocs_b_putmes(13, 0, CONSOLE_HEIGHT, CONSOLE_WIDTH - 1, modeline);
	_dos_c_curoff();
	oldbrk = _dos_breakck(2);
	curput(0, CONSOLE_INPUT);
	if (init) {
		printf("%s \033[31m" PROGNAME " " VERSION " " __DATE__ " by " AUTHOR "\033[m\n"
			   "%s \033[31mconnected to %s:%d\033[m\n",
			   getnow(), getnow(), irc_server, irc_port);
		if (irc_timeout > 0)
			printf("%s \033[32minactive timeout = %d second\033[m\n", getnow(), irc_timeout);
	}
}

/****************************************
	ʂ߂
****************************************/
void
tini_console()
{
	_iocs_b_consol(0, 0, CONSOLE_WIDTH - 1, CONSOLE_OHEIGHT - 1);
	_iocs_b_locate(0, CONSOLE_OHEIGHT - 1);
	_iocs_tpalet(0, BDCOLOR);
	_iocs_tpalet(3, FGCOLOR);
	_dos_c_curon();
	_dos_breakck(oldbrk);
}

/****************************************
	sʂɕ\
****************************************/
#define ACTION_FORMAT	"%s %c%s:%s%c \033[35m*** %s\033[m\n"
#define PRIVMSG_FORMAT	"%s %c%s:%s%c %s\n"
#define NOTICE_FORMAT	"%s %c%s:%s%c \033[37m%s\033[m\n"

static void
disp_mymessage(char *chan, const char *format, char *message)
{
	int ch = *chan;
	int left_mark, right_mark;

	left_mark = ch == '#' || ch == '&' ? '>' : '=',
	right_mark = ch == '#' || ch == '&' ? '<' : '=',
	printf(format, getnow(), left_mark, chan, irc_nick, right_mark, message);
}

/****************************************
	L[́EsҏW
****************************************/
static char onelinebuf[1024];

BOOLEAN
irc_console(int sock)
{
	static char *p = onelinebuf;
	static int count = 0;
	static time_t pre_input_time = 0;
	char linebuf[1024];
	char newbuf[1024];
	char chanbuf[1024];

	if (kbhit()) {
		int c;

		timebeep = FALSE;
		timeout = time(NULL);
		c = getch();
		if (ismbblead(c)) {
			*p++ = c;
			*p++ = getch();
			count += 2;
			*p = '\0';
		} else {
			switch (c) {
				case 0x01:	/* CTRL+A */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						if (*onelinebuf) {
							sjstr2jstr(newbuf, onelinebuf);
							sprintf(linebuf, "AWAY :%s\r\n", newbuf);
							*(p = onelinebuf) = '\0';
							count = 0;
						} else {
							strcpy(linebuf, "AWAY\r\n");
						}
#ifdef DEBUG
						if (debug)
							printf(linebuf);
#endif
						write(sock, linebuf, strlen(linebuf));
					}
					break;
				case 0x02:	/* CTRL+B */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						list_joined();
					}
					break;
				case 0x03:	/* CTRL+C */
					if (ctrlx == FALSE) {
						// list, names, who ̏o͂Jbg
						ignore_long_cmd = TRUE;
					} else {
						mdset("");
						irc_quit(sock, irc_quit_message);
					}
					break;
#ifdef DEBUG
				case 0x04:	/* CTRL+D */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						debug ^= TRUE;
						printf("%s \033[32mdebug %s\033[m\n", getnow(), debug ? "on" : "off");
					}
					break;
#endif
				case 0x08:	/* CTRL+H, BS */
					{
						char *tp;

						if (count == 0)
							return TRUE;
						for (tp = onelinebuf; *tp; tp++) {
							if (ismbblead(*tp)) {
								if (tp == onelinebuf + strlen(onelinebuf) - 2) {
									--count;
									--p;
									break;
								}
								tp++;
							}
						}
						--count;
						*--p = '\0';
					}
					break;
				case 0x09:	/* TAB, CTRL+I */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						list_ignore();
					} else {
						*p++ = ' ';
						count++;
						*p = '\0';
					}
					break;
				case 0x0a:	/* CTRL+J */
					if (current_channel >= 0) {
						if (ctrlx != FALSE) {
							ctrlx = FALSE;
							new_channel(current_channel);
							mdupdate();
						} else if (*onelinebuf) {
							int ch;

							sjchan2jchan(chanbuf, chanlist[current_channel].name);
							ch = *chanbuf;
							sjstr2jstr(newbuf, onelinebuf);
							disp_mymessage(chanlist[current_channel].name,
										   NOTICE_FORMAT,
										   onelinebuf);
							sprintf(linebuf, "NOTICE %s :%s\r\n", chanbuf, newbuf);
#ifdef DEBUG
							if (debug)
								printf(linebuf);
#endif
							write(sock, linebuf, strlen(linebuf));
							*(p = onelinebuf) = '\0';
							count = 0;
						}
					}
					break;
				case '\r':	/* CTRL+M */
					*p = '\0';	/* 󃊃^[h~ */
					/* flood control RFC1459 8.10 */
					if (*onelinebuf)
						while (time(NULL) - pre_input_time < 2)
							;
					mdset("");
					if (*onelinebuf != '/' && *onelinebuf) {
						if (current_channel >= 0) {
							int ch;

							sjchan2jchan(chanbuf, chanlist[current_channel].name);
							ch = *chanbuf;
							if (ctrlx == FALSE) {
								disp_mymessage(chanlist[current_channel].name,
											   PRIVMSG_FORMAT,
											   onelinebuf);
								sjstr2jstr(newbuf, onelinebuf);
								sprintf(linebuf, "PRIVMSG %s :%s\r\n", chanbuf, newbuf);
							} else {
								disp_mymessage(chanlist[current_channel].name,
											   ACTION_FORMAT,
											   onelinebuf);
								sjstr2jstr(newbuf, onelinebuf);
								sprintf(linebuf, "PRIVMSG %s :\001ACTION %s\001\r\n", chanbuf, newbuf);
							}
#ifdef DEBUG
							if (debug)
								printf(linebuf);
#endif
							write(sock, linebuf, strlen(linebuf));
						}
					} else if (*onelinebuf == '/') {
						irc_command(sock, onelinebuf + 1);
					}
					ctrlx = FALSE;
					*(p = onelinebuf) = '\0';
					count = 0;
					pre_input_time = time(NULL);
					break;
				case 0x0e:	/* CTRL+N */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						irc_names(sock);
					}
					break;
				case 0x10:	/* CTRL+P */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						if (current_channel >= 0) {
							int x;

							sjchan2jchan(chanbuf, chanlist[current_channel].name);
							x = *chanbuf;
							if (x == '#' || x == '&') {
								sprintf(linebuf, "PART %s\r\n", chanbuf);
#ifdef DEBUG
								if (debug)
									printf(linebuf);
#endif
								write(sock, linebuf, strlen(linebuf));
							} else {
								part_chanlist(chanbuf);
							}
						}
					}
					break;
				case 0x14:	/* CTRL+T */
					if (ctrlx != FALSE && current_channel >= 0) {
						ctrlx = FALSE;
						sjchan2jchan(chanbuf, chanlist[current_channel].name);
						if (*onelinebuf) {
							sjstr2jstr(newbuf, onelinebuf);
							sprintf(linebuf, "TOPIC %s :%s\r\n", chanbuf, newbuf);
							*(p = onelinebuf) = '\0';
							count = 0;
						} else {
							sprintf(linebuf, "TOPIC %s\r\n", chanbuf);
						}
#ifdef DEBUG
						if (debug)
							printf(linebuf);
#endif
						write(sock, linebuf, strlen(linebuf));
					}
					break;
				case 0x15:	/* CTRL+U */
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						if (*onelinebuf) {
							irc_info = xrealloc(irc_info, strlen(onelinebuf) + 1);
							strcpy(irc_info, onelinebuf);
							printf("%s \033[32muserinfo set '%s'\033[m\n", getnow(), irc_info);
						}
					}
					*(p = onelinebuf) = '\0';
					count = 0;
					break;
				case 0x18:	/* CTRL+X */
					if (ctrlx == FALSE) {
						ctrlx = TRUE;
						mdset("[CTRL-X]");
					} else {
						ctrlx = FALSE;
						mdset(onelinebuf);
						curput(count, CONSOLE_INPUT);
					}
					return TRUE;
				case 0x1a:	/* CTRL+Z */
					savelayout();
					tini_console();
					curput(count, CONSOLE_INPUT);
					system(NULL);
					init_console(FALSE);
					restorelayout();
					break;
				default:
					if (ctrlx != FALSE) {
						ctrlx = FALSE;
						mdset(onelinebuf);
						curput(count, CONSOLE_INPUT);
						return TRUE;
					}
					if (c < ' ' && c != '\a')
						return TRUE;
					*p++ = c;
					count++;
					*p = '\0';
					break;
			}
		}
		if (count >= CONSOLE_WIDTH) {
			count = CONSOLE_WIDTH - 1;
			p = onelinebuf + count;
			*p = '\0';
			if (ismbblead(*(p - 1))) {
				*--p = '\0';
				--count;
			}
			_dos_putchar('\a');
		}
		_iocs_b_putmes(3, 0, CONSOLE_INPUT, CONSOLE_WIDTH - 1, onelinebuf);
		curput(count, CONSOLE_INPUT);
	} else
		mdupdate();
	return TRUE;
}
