/*
 * ctcp.c
 */

__asm("	text\n	dc.b	'$Id: ctcp.c,v 1.2 1998/10/02 13:23:51 sharl Exp $'\n	even\n");

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "config.h"
#include "ctcp.h"
#include "command.h"
#include "channel.h"
#include "util.h"
#include "kanji.h"

/****************************************
	CTCP bZ[WsTu

	to:      JIS
	message: JIS
****************************************/
static void
ctcp_to_generic(int sock, char *to, char *command, char *message)
{
	char line[1024];

	if (message && *message)
		sprintf(line, "PRIVMSG %s :\001%s %s\001\r\n", to, command, message);
	else
		sprintf(line, "PRIVMSG %s :\001%s\001\r\n", to, command);
#ifdef DEBUG
	if (debug)
		printf(line);
#endif
	write(sock, line, strlen(line));
}

/****************************************
	CTCP bZ[Ws
****************************************/
void
ctcp_to(int sock, char *buf)
{
	char comm[1024] = {0};
	char to[1024] = {0}, to_jis[1024];
	char line[1024];
	int n;
	const CTCP *ctcp;

	sscanf(buf, "%*s %s %s%n", comm, to, &n);
	if (!*to || !*comm) {
		printf("\007\033[32musage: /ctcp command <nick|channel>\n"
			   "command: ACTION CLIENTINFO PING TIME USERINFO VERSION\033[m\n");
		return;
	}

	sjchan2jchan(to_jis, to);
	strupr(comm);
	ctcp = ctcp_word_set(comm, strlen(comm));
	if (ctcp != NULL) {
		switch (ctcp->cmd) {
			case CTCP_ACTION:
				{
					char newbuf[512], tmpbuf[512];

					sscanf(buf + n, " %[^\001]", tmpbuf);
					sjstr2jstr(newbuf, tmpbuf);
					printf(*to_jis == '#' || *to_jis == '&'
						   ? "%s >%s:%s< \033[35m*** %s\033[m\n"
						   : "%s =%s:%s= \033[35m*** %s\033[m\n",
						   getnow(), to, irc_nick, tmpbuf);
					ctcp_to_generic(sock, to_jis, comm, newbuf);
				}
				break;
			case CTCP_PING:
				irc_ping = time(NULL);
				ping_code = rand();
				sprintf(line, "PRIVMSG %s :\001PING %d\001\r\n", to, ping_code);
				write(sock, line, strlen(line));
#ifdef DEBUG
				if (debug)
					printf("ping_code = %d\n", ping_code);
#endif
				break;
			case CTCP_CLIENTINFO:
			case CTCP_TIME:
			case CTCP_USERINFO:
			case CTCP_VERSION:
				ctcp_to_generic(sock, to_jis, comm, NULL);
				break;
			default:
				goto err;
				break;
		}
	} else {
err:
		printf("%s \033[36mCTCP %s: command error\033[m\n", getnow(), comm);
	}
}

/****************************************
	CTCP vւ̕ԓTu
****************************************/
static void
ctcp_notice_sub(int sock, char *nick, char *comm, char *message, BOOLEAN supported)
{
	char line[1024];

	if (supported) {
		printf("%s \033[31mCTCP %s from %s\033[m\n", getnow(), comm, nick);
		sprintf(line, "NOTICE %s :\001%s %s\001\r\n", nick, comm, message);
	} else {
		printf("%s \033[31mCTCP %s (unsupported) from %s [%s]\033[m\n",
			   getnow(), comm, nick, message);
		sprintf(line, "NOTICE %s :\001ERRMSG :Unrecognized command: '%s'\001\r\n", nick, comm);
	}
#ifdef DEBUG
	if (debug)
		printf(line);
#endif
	write(sock, line, strlen(line));
}

/****************************************
	̃[U CTCP vւ̕ԓ
****************************************/
void
ctcp_from(int sock, char *nick, char *chan, char *buf)
{
	char line[1024];
	char newbuf[1024];
	char comm[512];
	int n;
	const CTCP *ctcp;

	sscanf(buf, "\001%[^ \001]%n", comm, &n);
	strupr(comm);
	ctcp = ctcp_word_set(comm, strlen(comm));
	if (ctcp) {
		switch (ctcp->cmd) {
			case CTCP_ACTION:
				{
					char tmpbuf[512];

					sscanf(buf, "%*s %[^\001]", tmpbuf);
					jstr2sjstr(newbuf, tmpbuf);
					printf(*chan == '#' || *chan == '&' || *chan == '%'
						   ? "%s <%s:%s> \033[35m*** %s\033[m\n"
						   : "%s =%s:%s= \033[35m*** %s\033[m\n",
						   getnow(), chan, nick, newbuf);
				}
				break;
			case CTCP_CLIENTINFO:
				ctcp_notice_sub(sock, nick, comm,
								":ACTION CLIENTINFO PING TIME USERINFO VERSION", TRUE);
				break;
			case CTCP_PING:
				sscanf(buf + n, " %s", newbuf);
				ctcp_notice_sub(sock, nick, comm, newbuf, TRUE);
				break;
			case CTCP_TIME:
				{
					time_t tim;
					struct tm *tm;
					char timbuf[256];

					tim = time(NULL);
					tm = localtime(&tim);
					strftime(timbuf, sizeof(timbuf), "%a %b %d %T %Y", tm);
					ctcp_notice_sub(sock, nick, comm, timbuf, TRUE);
				}
				break;
			case CTCP_USERINFO:
				sprintf(line, ":%s", irc_info);
				sjstr2jstr(newbuf, line);
				ctcp_notice_sub(sock, nick, comm, newbuf, TRUE);
				break;
			case CTCP_VERSION:
				sjstr2jstr(newbuf, PROGNAME " " VERSION " (X68k; " __DATE__ ")");
				ctcp_notice_sub(sock, nick, comm, newbuf, TRUE);
				break;
#ifdef DCC_TEST
			case CTCP_DCC:
				{
					char type[512], argument[512], address[512], port[512];
					int n;

// 01:31 DCC SEND from X-K6
// 01:31 argument[gusa.txt] address[3534610728] port[1126] size[ 23\001]
// 01:35 DCC CHAT from X-K6
// 01:35 argument[chat] address[3534610728] port[1128\001]

					sscanf(buf, "%*s %s %s %s %s%n", type, argument, address, port, &n);
					printf("%s \033[35mDCC %s from %s\033[m\n", getnow(), type, nick);
					printf("%s argument[%s] address[%s] port[%s]",
						   getnow(), argument, address, port);
					if (*(buf + n))
						printf(" size[%s]\n", buf + n);
					else
						printf("\n");
				}
				break;
#endif
			default:
				goto err;
				break;
		}
	} else {
err:
		ctcp_notice_sub(sock, nick, comm, buf, FALSE);
	}
}

/****************************************
	CTCP NGXǧʕ\
****************************************/
void
ctcp_action(char *nick, char *buf)
{
	int len;
	char comm[512], newbuf[512];

	sscanf(buf, "\001%s%n", comm, &len);
	if (!strcasecmp(comm, "PING")) {
		time_t ping;

		sscanf(buf + len, " %d", &ping);
		if (ping == ping_code) {
			printf("%s \033[31m*** %s(%s) = %d second\033[m\n",
				   getnow(), comm, nick, time(NULL) - irc_ping);
		} else {
			printf("\033[32mIllegal PING code returns from %s(%d)\033[m\n",
				   nick, ping);
		}
	} else {
		char tmpbuf[1024];

		sscanf(buf + len + 1, " %[^\001]", newbuf);
		jstr2sjstr(tmpbuf, newbuf);
		printf("%s \033[31m*** %s(%s) = %s\033[m\n",
			   getnow(), comm, nick, *tmpbuf == ':' ? tmpbuf + 1 : tmpbuf);
	}
}
