/*
 * main.c
 */

__asm("	text\n	dc.b	'$Id: main.c,v 1.27 1998/11/03 15:24:28 sharl Exp $'\n	even\n");

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <time.h>
#include <err.h>

#include "getopt.h"
#include "config.h"
#include "util.h"
#include "socket.h"
#include "console.h"
#include "command.h"
#include "channel.h"

char *__progname;
#ifdef DEBUG
BOOLEAN debug = FALSE;
#endif

static struct option long_opt[] = {
#ifdef DEBUG
	{"debug", no_argument, NULL, 'd'},
#endif
	{"help", no_argument, NULL, 'h'},
	{"no-join", no_argument, NULL, 'N'},
	{"nick", required_argument, NULL, 'n'},
	{"port", required_argument, NULL, 'p'},
	{"supprt", no_argument, NULL, 's'},
	{"timeout", required_argument, NULL, 't'},
	{"version", no_argument, NULL, 'v'},
	{NULL}
};

/****************************************
	gp@
****************************************/
static volatile void
usage()
{
	fprintf(stderr, "usage: " PROGNAME " [options] [server name] [port]\n"
	       "options:\n"
#ifdef DEBUG
	       "-d, --debug		debug mode on\n"
#endif
	       "-h, --help		show this\n"
	       "-N, --no-join		don't join to initial channel\n"
	       "-n, --nick nick		set nick\n"
	       "-p, --port port		set port number\n"
	       "-s, --support		show keyword ." PROGNAME "rc supported\n"
	       "-t, --timeout sec	set inactive timeout seconds\n"
	       "-v, --version		show version\n");
	exit(EXIT_FAILURE);
}

/****************************************
	C
****************************************/
int
main(int argc, char *argv[])
{
	int c;
	int sock;
	BOOLEAN retry = FALSE;
	BOOLEAN port_flag = FALSE;

	__progname = argv[0];

	if (ReadConfig() < 0)
		return EXIT_FAILURE;

	while ((c = getopt_long(argc, argv, "dhNn:p:st:v", long_opt, 0)) != EOF) {
		switch (c) {
			case 'd':
#ifdef DEBUG
				debug++;
#endif
				break;
			case 'h':
				usage();
			case 'N':
				nojoin = TRUE;
				break;
			case 'n':
				irc_nick = xrealloc(irc_nick, strlen(optarg) + 1);
				strcpy(irc_nick, optarg);
				break;
			case 'p':
				irc_port = atol(optarg);
				port_flag = TRUE;
				break;
			case 's':
				{
					DIREC *p;

					for (p = directive; p->name; p++)
						printf("%s\n", p->name);
				}
				return EXIT_SUCCESS;
			case 't':
				irc_timeout = atol(optarg);
				break;
			case 'v':
				fprintf(stderr, PROGNAME " " VERSION " (C)1998 " AUTHOR "\n");
				exit(EXIT_FAILURE);
			default:
				usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* set irc_server from command line */
	if (argc > 0) {
		irc_server = xrealloc(irc_server, strlen(argv[0]) + 1);
		strcpy(irc_server, argv[0]);
		if (!port_flag)
			irc_port = DEFAULT_PORT;
	}
	argc--;
	argv++;
	/* set irc_port from command line */
	if (argc > 0)
		irc_port = atol(argv[0]);

#ifdef DEBUG
	if (debug) {
		int i;
		char **p;

		printf("server  %s\n", irc_server);
		printf("pass    %s\n", irc_pass);
		printf("port    %d\n", irc_port);
		printf("nick    %s\n", irc_nick);
		printf("mail    %s\n", irc_user);
		printf("name    %s\n", irc_name);
		printf("info    %s\n", irc_info);
		printf("quit    %s\n", irc_quit_message);
		printf("keyword %s\n", irc_keyword);
		printf("bell    %s\n", irc_keyword_bell);
		printf("bold    %s\n", irc_keyword_bold);
		for (i = 0; i < IRC_JOIN_MAX && irc_channel[i].name; i++)
			printf("chan    %s,%s\n", irc_channel[i].name, irc_channel[i].key);
		if (irc_autonick)
			for (p = irc_autonick; *p; p++)
				printf("anick   %s\n", *p);
		if (irc_autochanops)
			for (p = irc_autochanop; *p; p++)
				printf("achop   %s\n", *p);
		if (irc_isons)
			for (p = irc_ison; *p; p++)
				printf("ison    %s\n", *p);
		printf("ison/s  %d\n", irc_ison_interval);
		if (irc_ignore) {
			IGNORE *ip;

			for (ip = irc_ignore; ip; ip = ip->next)
				if (ip->on)
					printf("ignore  %s\n", ip->nick);
		}
		if (irc_timeout > 0)
			printf("timeout %d\n", irc_timeout);
		printf("aquit   %s\n", irc_autoquit_message);
	}
#endif

	{
		int err = 0;

		if (!irc_server || !*irc_server)
			fprintf(stderr, RC_SERVER " is NOT set.\n"), err++;
		if (!irc_nick || !*irc_nick)
			fprintf(stderr, RC_NICK " is NOT set.\n"), err++;
		if (!irc_user || !*irc_user)
			fprintf(stderr, RC_MAILADDRESS " is NOT set.\n"), err++;
		if (!irc_name || !*irc_name)
			fprintf(stderr, RC_NAME " is NOT set.\n"), err++;
		if (err)
			return EXIT_FAILURE;
	}

 l_retry:
	if ((sock = irc_connect()) < 0)
		return EXIT_FAILURE;
	if (!retry)
		init_console(TRUE);
	if (irc_identify(sock)) {
		time_t t_ison;

		// user mode ݒ
		if (irc_usermode) {
			char linbuf[512];

			sprintf(linbuf, "MODE %s %s\r\n", irc_nick, irc_usermode);
#ifdef DEBUG
			if (debug)
				printf(linbuf);
#endif
			write(sock, linbuf, strlen(linbuf));
		}
		// `l JOIN
		irc_initial_join(sock);
		// ISON sĂ܂
		if (irc_isons) {
			ison = TRUE;
#ifdef DEBUG
			if (debug)
				printf(irc_ison_command);
#endif
			write(sock, irc_ison_command, strlen(irc_ison_command));
		}

		t_ison = timeout = time(NULL);
		while (1) {
			int len;
			time_t tim;
			static BOOLEAN autoquit = FALSE;

			tim = time(NULL);
			if (irc_timeout > 0 && !autoquit && tim - timeout > irc_timeout) {
				autoquit = TRUE;
				irc_quit(sock, irc_autoquit_message);
			}
			if ((len = socklen(sock, 0)) > 0)
				if (irc_read(sock) == FALSE)
					break;
			if (irc_console(sock) == FALSE)
				break;
			if (irc_isons != NULL && ison == FALSE
				&& t_ison != tim && tim % irc_ison_interval == 0) {
				t_ison = tim;
				ison = TRUE;
#ifdef DEBUG
				if (debug)
					printf("%s", irc_ison_command);
#endif
				write(sock, irc_ison_command, strlen(irc_ison_command));
			}
		}
	}

	if (irc_autoreconnect && !explicitly_quit && !strcmp(sockstate(sock), "CLOSE WAIT")) {
		printf("%s \033[32mreconnect to %s:%d\033[m\n", getnow(), irc_server, irc_port);
		retry = TRUE;
		goto l_retry;
	}

	irc_disconnect(sock);

	tini_console();

	return EXIT_SUCCESS;
}
