/*
 * socket.c
 */

__asm("	text\n	dc.b	'$Id: socket.c,v 1.14 1998/11/03 15:25:21 sharl Exp $'\n	even\n");

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#define __ARPA_INLINE__
#include <arpa/nameser.h>
#include <netdb.h>
#include <err.h>

#include "config.h"
#include "socket.h"
#include "irc.h"
#include "util.h"
#include "kanji.h"

/****************************************
	ڑ
	s -1
****************************************/
int
irc_connect()
{
	struct sockaddr_in saddr;
	struct hostent *shost;
	int sock;

	memset(&saddr, 0, sizeof(saddr));
	if (isipaddr(irc_server)) {
		saddr.sin_addr.s_addr = inet_addr(irc_server);
		shost = gethostbyaddr((char *)&saddr.sin_addr, sizeof(saddr.sin_addr), AF_INET);
		if (shost)
			irc_server = shost->h_name;
		else {
			fprintf(stderr, "%s: gethostbyaddr() failed.\n", irc_server);
			return -1;
		}
	} else {
		shost = gethostbyname(irc_server);
		if (shost) {
			memcpy(&saddr.sin_addr, shost->h_addr, shost->h_length);
		} else {
			fprintf(stderr, "%s: host name lookup failed.\n", irc_server);
			return -1;
		}
	}
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(irc_port);
 	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		fprintf(stderr, "Can't create socket.\n");
		return -1;
	}
#ifdef DEBUG
	if (debug)
		printf("socket OK\n");
#endif
	if (connect(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		fprintf(stderr, "Unable to connect: %s\n", sockerr(sock));
		return -1;
	}
#ifdef DEBUG
	if (debug)
		printf("connect OK\n");
#endif

	return sock;
}

/****************************************
	ؒf
****************************************/
int
irc_disconnect(int sock)
{
	if (shutdown(sock, 2) < 0)
		errx(EXIT_FAILURE, "shutdown failed.\n");
	close(sock);
	printf("%s \033[32mConnection closed.\033[m\n", getnow());
	return 0;
}

/****************************************
	[UF؃Tu
****************************************/
static BOOLEAN
irc_ident_sub(int sock, char *nick, BOOLEAN retry)
{
	size_t len;
	char buf[BUFSIZ];
	char newbuf[BUFSIZ];
	char tmpbuf[BUFSIZ];

	if (irc_pass) {
		sprintf(buf, "PASS %s\r\n", irc_pass);
#ifdef DEBUG
		if (debug)
			puts(buf);
#endif
		write(sock, buf, strlen(buf));
	}
	sprintf(buf, "NICK %s\r\n", nick);
#ifdef DEBUG
	if (debug)
		printf(buf);
#endif
	write(sock, buf, strlen(buf));

	if (!retry) {
		sjstr2jstr(newbuf, irc_name);
		sprintf(buf, "USER %s * %s :%s\r\n", irc_user, irc_server, newbuf);
#ifdef DEBUG
		if (debug)
			printf(buf);
#endif
		write(sock, buf, strlen(buf));
	}

	while (socklen(sock, 0) == 0)
		;
	if ((len = recvline(sock, tmpbuf, BUFSIZ - 1)) > 0) {
		*(tmpbuf + len) = '\0';
		chop_crlf(tmpbuf);
		return irc_parseline(sock, tmpbuf);
	}
	return TRUE;
}

/****************************************
	[UF؃C
****************************************/
BOOLEAN
irc_identify(int sock)
{
	char **nick;

	if (irc_ident_sub(sock, irc_nick, FALSE))
		return TRUE;
	if (irc_autonick)
		for (nick = irc_autonick; *nick; nick++) {
			printf("%s \033[32mRetry nickname '%s'.\033[m\n", getnow(), *nick);
			if (irc_ident_sub(sock, *nick, TRUE)) {
				irc_nick = *nick;
				return TRUE;
			}
		}
	return FALSE;
}

/****************************************
****************************************/
static BOOLEAN
checkPING(int sock, char *buf)
{
	if (!strncmp(buf, "PING", 4)) {
#ifdef DEBUG
		if (debug)
			puts(buf);
#endif
		*(buf + 1) = 'O';	/* PING -> PONG */
#ifdef DEBUG
		if (debug)
			puts(buf);
#endif
		write(sock, buf, strlen(buf));
		return TRUE;
	}
	return FALSE;
}

/****************************************
	G[`FbN
****************************************/
static BOOLEAN
checkERROR(char *buf)
{
	if (!strncmp(buf, "ERROR", 5)) {
//ERROR :Closing Link: Sharl[~sharl@cs11804.ppp.infoweb.or.jp] (BL@F|$N$?$a$K?2$k$Y$7B)
		char reason[256], account[256], message[256], mesbuf[256];

		sscanf(buf, "%*s :%[^:]: %[^[]%*s %[^\r\n]", reason, account, message);
		if (!strcmp(reason, "Closing Link")) {
			jstr2sjstr(mesbuf, message);
			printf("%s \033[35m%s has left IRC %s\033[m\n", getnow(), account, mesbuf);
		} else {
			puts(buf);
		}
		return TRUE;
	}
	return FALSE;
}

/****************************************
	\Pbgsǂݍ
****************************************/
BOOLEAN
irc_read(int sock)
{
	size_t len;
	char buf[BUFSIZ];

	/* s͂₷悤Ɉsǂݍ */
	if ((len = recvline(sock, buf, BUFSIZ - 1)) > 0) {
		*(buf + len) = '\0';
		chop_crlf(buf);

		if (checkPING(sock, buf))
			return TRUE;
		if (checkERROR(buf))
			return FALSE;
		irc_parseline(sock, buf);
		return TRUE;
	}
	return FALSE;
}
